//+------------------------------------------------------------------+
//|                                                   ListSample.mq5 |
//|                        Copyright 2010, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2010, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"
#property script_show_inputs
//--- include
#include <Arrays\List.mqh>
//--- defines
enum MY_COMPARE_AND_SORT_MODES
  {
   MY_MODE_BY_ALL,
   MY_MODE_BY_KEY1,
   MY_MODE_BY_KEY2,
   MY_MODE_BY_KEY1_KEY2,
   MY_MODE_BY_KEY2_KEY1
  };
//--- inputs
input int InpMatrixSizeX=10;
input int InpMatrixSizeY=10;
//+------------------------------------------------------------------+
//| Class CMyListElement.                                            |
//| Purpose: Class implementing list elements.                       |
//|          Derives from the CObject class.                         |
//+------------------------------------------------------------------+
class CMyListElement : public CObject
  {
protected:
   uint              m_key1;           // key field
   uint              m_key2;           // key field
   int               m_value;          // data

public:
                     CMyListElement();
   //--- Methods of data access
   uint              Key1()           { return(m_key1);  }
   void              Key1(uint key)   { m_key1=key;      }
   uint              Key2()           { return(m_key2);  }
   void              Key2(uint key)   { m_key2=key;      }
   int               Value()          { return(m_value); }
   void              Value(int value) { m_value=value;   }
   void              SetData(uint key1,uint key2,int value);
   //--- Methods of working with files
   virtual bool      Save(int file_handle);
   virtual bool      Load(int file_handle);
protected:
   virtual int       Compare(const CObject *object,int mode);
  };
//+------------------------------------------------------------------+
//| Constructor of the CMyListElement class.                         |
//| INPUT:  None.                                                    |
//| OUTPUT: None.                                                    |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
void CMyListElement::CMyListElement()
  {
//--- Initialization of user data
   m_key1 =0;
   m_key2 =0;
   m_value=0;
  }
//+------------------------------------------------------------------+
//| Fill method.                                                     |
//| INPUT:  key1  - New value for the m_key1 field,                  |
//|         key2  - New value for the m_key2 field,                  |
//|         value - New value for the m_value field.                 |
//| OUTPUT: None.                                                    |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
void CMyListElement::SetData(uint key1,uint key2,int value)
  {
   m_key1 =key1;
   m_key2 =key2;
   m_value=value;
  }
//+------------------------------------------------------------------+
//| Writing the list element to file.                                |
//| INPUT:  file_handle - handle of file opened earlier for writing. |
//| OUTPUT: True if Ok, otherwise false.                             |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
bool CMyListElement::Save(int file_handle)
  {
//--- Checks
   if(file_handle<0) return(false);
//--- Writing user data
//--- Writing user variable m_key1
   if(FileWriteInteger(file_handle,m_key1,INT_VALUE)!=INT_VALUE)  return(false);
//--- Writing user variable m_key2
   if(FileWriteInteger(file_handle,m_key2,INT_VALUE)!=INT_VALUE)  return(false);
//--- Writing user variable m_value
   if(FileWriteInteger(file_handle,m_value,INT_VALUE)!=INT_VALUE) return(false);
//---
   return(true);
  }
//+------------------------------------------------------------------+
//| Reading the list element from the file.                          |
//| INPUT:  file_handle - handle of file opened earlier for reading. |
//| OUTPUT: True if Ok, otherwise false.                             |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
bool CMyListElement::Load(int file_handle)
  {
//--- Checks
   if(file_handle<0) return(false);
//--- Reading
   if(FileIsEnding(file_handle)) return(false);
//--- Reading the user variable m_key1
   m_key1=(uint)FileReadInteger(file_handle,CHAR_VALUE);
//--- Reading the user variable m_key2
   m_key2=(uint)FileReadInteger(file_handle,SHORT_VALUE);
//--- Reading the user variable m_value
   m_value=(int)FileReadInteger(file_handle,INT_VALUE);
//---
   return(true);
  }
//+------------------------------------------------------------------+
//| Compare to another element of the list by algorithm.             |
//| INPUT:  object - objet instance to compare,                      |
//|         mode   - ID of algorithm to compare.                     |
//| OUTPUT: Result of comparison (>0,0,<0).                          |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
int CMyListElement::Compare(const CObject *object,int mode)
  {
   int res=0;
//--- No need to compare the object pointer, because call is made from method Sort
   //--- Explicit type casting
   CMyListElement *obj=object;
//---
   switch(mode)
     {
      //--- Compare all user data
      //--- Used by function of comaring lists
      case MY_MODE_BY_ALL:
         res=(int)(m_key1-obj.m_key1);
         if(res==0)
            res=(int)(m_key2-obj.m_key2);
         //--- To get a certain result compare unsigned numbers
         if(res==0)
            res=(int)((uint)m_value-(uint)obj.m_value);
         break;
       //--- Compare by m_key1
       case MY_MODE_BY_KEY1:
         res=(int)(m_key1-obj.m_key1);
         break;
       //--- Compare by m_key2
       case MY_MODE_BY_KEY2:
         res=(int)(m_key2-obj.m_key2);
         break;
       //--- Compare by m_key1 and m_key2
       case MY_MODE_BY_KEY1_KEY2:
         res=(int)(m_key1-obj.m_key1);
         if(res==0)
            res=(int)(m_key2-obj.m_key2);
         break;
       //--- Compare by m_key2 and m_key1
       case MY_MODE_BY_KEY2_KEY1:
         res=(int)(m_key2-obj.m_key2);
         if(res==0)
            res=(int)(m_key1-obj.m_key1);
         break;
       default:
         break;
     }
//---
   return(res);
  }
//+------------------------------------------------------------------+
//| Class CMyList.                                                   |
//| Purpose: Class implementing sample list.                         |
//|          Derives from the CList class.                           |
//+------------------------------------------------------------------+
class CMyList : public CList
  {
public:
   //--- Methods of search through list by user data
   CMyListElement*   SearchByKey1(uint key1);
   CMyListElement*   SearchByKey2(uint key2);
   CMyListElement*   SearchByKey1Key2(uint key1,uint key2);
   CMyListElement*   SearchByKey2Key1(uint key2,uint key1);
   //--- Method of list element creation
   virtual CObject  *CreateElement();
  };
//+------------------------------------------------------------------+
//| Search for element in the list by key1.                          |
//| INPUT:  key1 - desired value of key1.                            |
//| OUTPUT: Pointer of the found element of list, or NULL.           |
//| REMARK: If list isn't sorted by field key1                       |
//|         (MY_MODE_BY_KEY1), do not search.                        |
//+------------------------------------------------------------------+
CMyListElement* CMyList::SearchByKey1(uint key1)
  {
   CMyListElement *res=NULL;
   CMyListElement *node;
//---
   if(!m_data_sort || m_sort_mode!=MY_MODE_BY_KEY1)
     {
      //--- The list is not sorted
      return(NULL);
     }
//--- Search in the sorted array
   node=new CMyListElement;
   if(node==NULL) return(NULL);
   node.Key1(key1);
   res=Search(node);
   delete node;
//--- Return search results
   return(res);
  }
//+------------------------------------------------------------------+
//| Search for element in the list by key2.                          |
//| INPUT:  key2 - desired value of key2.                            |
//| OUTPUT: Pointer of the found element of list, or NULL.           |
//| REMARK: If list isn't sorted by field key2                       |
//|         (MY_MODE_BY_KEY2), do not search.                        |
//+------------------------------------------------------------------+
CMyListElement* CMyList::SearchByKey2(uint key2)
  {
   CMyListElement *res=NULL;
   CMyListElement *node;
//---
   if(!m_data_sort || m_sort_mode!=MY_MODE_BY_KEY2)
     {
      //--- The list is not sorted
      return(NULL);
     }
//--- Search in the sorted array
   node=new CMyListElement;
   if(node==NULL) return(NULL);
   node.Key2(key2);
   res=Search(node);
   delete node;
//--- Return search results
   return(res);
  }
//+------------------------------------------------------------------+
//| Search for element in the list by key1 and key2.                 |
//| INPUT:  key1 - desired value of key1,                            |
//| INPUT:  key2 - desired value of key2.                            |
//| OUTPUT: Pointer of the found element of list, or NULL.           |
//| REMARK: If list isn't sorted by fields key1 and key2             |
//|         (MY_MODE_BY_KEY1_KEY2), do not search.                   |
//+------------------------------------------------------------------+
CMyListElement* CMyList::SearchByKey1Key2(uint key1,uint key2)
  {
   CMyListElement *res=NULL;
   CMyListElement *node;
//---
   if(!m_data_sort || m_sort_mode!=MY_MODE_BY_KEY1_KEY2)
     {
      //--- The list is not sorted
      return(NULL);
     }
//--- Search in the sorted array
   node=new CMyListElement;
   if(node==NULL) return(NULL);
   node.Key1(key1);
   node.Key2(key2);
   res=Search(node);
   delete node;
//--- Return search results
   return(res);
  }
//+------------------------------------------------------------------+
//| Search for element in the list by key2 and key1.                 |
//| INPUT:  key2 - desired value of key2,                            |
//|         key1 - desired value of key1.                            |
//| OUTPUT: Pointer of the found element of list, or NULL.           |
//| REMARK: If list isn't sorted by fields key2 and key1             |
//|         (MY_MODE_BY_KEY2_KEY1), do not search.                   |
//+------------------------------------------------------------------+
CMyListElement* CMyList::SearchByKey2Key1(uint key2,uint key1)
  {
   CMyListElement *res=NULL;
   CMyListElement *node;
//---
   if(!m_data_sort || m_sort_mode!=MY_MODE_BY_KEY2_KEY1)
     {
      //--- The list is not sorted
      return(NULL);
     }
//--- Search in the sorted array
   node=new CMyListElement;
   if(node==NULL) return(NULL);
   node.Key1(key1);
   node.Key2(key2);
   res=Search(node);
   delete node;
//--- Return search results
   return(res);
  }
//+------------------------------------------------------------------+
//| Create a new element for the list.                               |
//| INPUT:  None.                                                    |
//| OUTPUT: Pointer to the new list element of OK, otherwise NULL.   |
//| REMARK: None.                                                    |
//+------------------------------------------------------------------+
CObject *CMyList::CreateElement()
  {
   CMyListElement *node=new CMyListElement;
//---
   return(node);
  }
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
   int             i,j;
   CMyList         list; // Instance of class CMyList 
   CMyListElement *node; // Pointer to the instance of class CMyListElement 
//--- Fill list with instances of class CMyListElement in the amount of InpMatrixSizeX*InpMatrixSizeY.
   MathSrand(0);
   for(i=0;i<InpMatrixSizeX;i++)
      for(j=0;j<InpMatrixSizeY;j++)
        {
         node=new CMyListElement;
         if(node==NULL)
           {
            printf(__FUNCTION__+": create error");
            return;
           }
         node.Key1(i);
         node.Key2(j);
         node.Value(MathRand());
         //--- Add element to list (will be deleted automatically))
         if(list.Add(node)==-1)
           {
            delete node;
            printf(__FUNCTION__+": add error [%d,%d]",i,j);
            return;
           }
         else
            printf("add element[%d,%d]=%d",i,j,node.Value());
        }
//--- Sort the list
   list.Sort(MY_MODE_BY_KEY2_KEY1);
   for(i=0;i<InpMatrixSizeY;i++)
      for(j=0;j<InpMatrixSizeX;j++)
        {
         node=list.SearchByKey2Key1(i,j);
         if(node==NULL)
           {
            printf(__FUNCTION__+": search error [%d,%d]",i,j);
            return;
           }
         else
            printf("search result: element[%d,%d]=%d",i,j,node.Value());
        }
  }
//+------------------------------------------------------------------+
